/****************************************************************************

  This file is part of the VRender library.
  Copyright (C) 2005 Cyril Soler (Cyril.Soler@imag.fr)
  Version 1.0.0, released on June 27, 2005.

  http://artis.imag.fr/Members/Cyril.Soler/VRender

  VRender is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation; either version 2 of the License, or (at your option) any later
  version.
  
  VRender is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with
  VRender; if not, write to the Free Software Foundation, Inc., 59 Temple
  Place, Suite 330, Boston, MA  02111-1307  USA

*****************************************************************************/

#ifndef _VRENDER_VECTOR2_H
#define _VRENDER_VECTOR2_H

#include <stdexcept>
#include <iostream>

namespace vrender
{
	template<class T> class T_Vector2
	{
		public:
			// ---------------------------------------------------------------------------
			//! @name Constant
			//@{
			static const T_Vector2 null;
			static const T_Vector2 inf;
			//@}

			// ---------------------------------------------------------------------------
			//! @name Constructor(s) and destructor
			//@{
			T_Vector2 ();
			~T_Vector2 ();
			T_Vector2 (const T_Vector2<T>&);
			T_Vector2 (T,T);
			// I don't like this at all, because it allows things like 
			// T_NVector2(Point3), which we certainly don't want although the compiler
			// won't complain at all about it. This produces bugs very hard to find
			// (I got one by doing 
			//       T_NVector2 toto(ray.getOrigin()) 
			// instead of
			//       Point3 toto(ray.getOrigin()) 
			//
			template <class U> inline explicit T_Vector2 (const U&v)
			{
				_xyz[0] = v[0] ;
				_xyz[1] = v[1] ;
			}
			//@}

			// ---------------------------------------------------------------------------
			//! @name Access methods
			//@{
			inline T  x() const { return _xyz[0]; }
			inline T  y() const { return _xyz[1]; }
			inline void  setX(T r) { _xyz[0] = r; }
			inline void  setY(T r) { _xyz[1] = r; }
			inline void  setXY (T x,T y) { _xyz[0] = x; _xyz[1] = y; }
			//@}

			// ---------------------------------------------------------------------------
			//! @name Assignment
			//@{
			inline T_Vector2& operator= (const T_Vector2<T>& u)  { _xyz[0] = u._xyz[0]; _xyz[1] = u._xyz[1]; return *this; }
			//@}
#ifdef A_VIRER
			// ---------------------------------------------------------------------------
			//! @name Comparisons
			//@{
			template<class U> friend bool operator== (const T_Vector2<U>&,const T_Vector2<U>&);
			template<class U> friend bool operator!= (const T_Vector2<U>&,const T_Vector2<U>&);
			bool isEqualTo (const T_Vector2<T>&,T epsilon=Real_epsilon) const;
			bool isNull (T epsilon=Real_epsilon) const;
			//@}
#endif

			// ---------------------------------------------------------------------------
			//! @name Algebraic operations
			//@{
			inline T_Vector2<T>& operator+= (const T_Vector2<T>& v)
			{ 
				_xyz[0] += v._xyz[0]; 
				_xyz[1] += v._xyz[1]; 
				return *this;
			}

			inline T_Vector2<T>& operator-= (const T_Vector2<T>& v)
			{ 
				_xyz[0] -= v._xyz[0]; 
				_xyz[1] -= v._xyz[1]; 
				return *this;
			}

			inline T_Vector2<T>& operator*= (T f) { _xyz[0] *= f; _xyz[1] *= f; return *this;}
			inline T_Vector2<T>& operator/= (T f) { _xyz[0] /= f; _xyz[1] /= f; return *this;}

			template<class U> friend T_Vector2<U> operator- (const T_Vector2<U>&);
			template<class U> friend T_Vector2<U> min(const T_Vector2<U>&,const T_Vector2<U>&) ;
			template<class U> friend T_Vector2<U> max(const T_Vector2<U>&,const T_Vector2<U>&) ;

			inline T_Vector2<T> operator+(const T_Vector2<T>& u) const 
			{ 
				return T_Vector2<T>(_xyz[0]+u._xyz[0],_xyz[1]+u._xyz[1]); 
			}
			inline T_Vector2<T> operator-(const T_Vector2<T>& u) const 
			{ 
				return T_Vector2<T>(_xyz[0]-u._xyz[0],_xyz[1]-u._xyz[1]); 
			}

			inline T    operator*(const T_Vector2<T>& u) const 
			{ 
				return _xyz[0]*u._xyz[0] + _xyz[1]*u._xyz[1] ; 
			}

			inline T operator^(const T_Vector2<T>& v) const 
			{ 
				return _xyz[0]*v._xyz[1] - _xyz[1]*v._xyz[0] ; 
			}

			T_Vector2<T> operator/ (T v) { return T_Vector2<T>(_xyz[0]/v,_xyz[1]/v); }
			T_Vector2<T> operator* (T v) { return T_Vector2<T>(_xyz[0]*v,_xyz[1]*v); }

			template<class U> friend T_Vector2<U> operator* (U,const T_Vector2<U>&);
			//@}

			// ---------------------------------------------------------------------------
			//! @name Metrics
			//@{
			T norm       () const;
			T squareNorm () const;
			T infNorm    () const; /// Should be used for most comparisons, for efficiency reasons.
			//@}
			// ---------------------------------------------------------------------------
			//! @name Stream overrides
			//@{
			template<class U> friend std::ostream& operator<< (std::ostream&,const T_Vector2<U>&);
			//@}

			T  operator[] (int i) const 
			{ 
				if((i < 0)||(i > 1))
					throw std::runtime_error("Out of bounds in T_Vector2::operator[]") ;

				return _xyz[i]; 
			}

			T& operator[] (int i) 
			{ 
				if((i < 0)||(i > 1))
					throw std::runtime_error("Out of bounds in T_Vector2::operator[]") ;

				return _xyz[i]; 
			}

		private:
			T _xyz[2];  //!< The 3 vector components

	}; // interface of T_Vector2


	// -------------------------------------------------------------------------- //
	//                                                                            //
	//  CLASS IMPLEMENTATION                                                      //
	//                                                                            //
	// -------------------------------------------------------------------------- //
	// -----------------------------------------------------------------------------
	//! Default constructor
	template<class T> T_Vector2<T>::T_Vector2 ()
	{
		_xyz[0] = 0.0;
		_xyz[1] = 0.0;
	}

	// -----------------------------------------------------------------------------
	//! Default destructor
	template<class T> T_Vector2<T>::~T_Vector2 ()
	{
	}

	// -----------------------------------------------------------------------------
	//! Copy constructor
	template<class T> T_Vector2<T>::T_Vector2 (const T_Vector2& u)
	{
		setXY(u[0],u[1]);
	}

	// -----------------------------------------------------------------------------
	//! Create a vector from real values
	template<class T> T_Vector2<T>::T_Vector2 (T x,T y)
	{
		setXY(x,y);
	}


	// -----------------------------------------------------------------------------
	//! Inverse
	template<class T> T_Vector2<T> operator- (const T_Vector2<T>& u)
	{
		return T_Vector2<T>(-u._xyz[0], -u._xyz[1]) ;
	}


	// -----------------------------------------------------------------------------
	//! Left multiplication by a real value
	template<class T> T_Vector2<T> operator* (T r,const T_Vector2<T>& u)
	{
		return T_Vector2<T>(r*u._xyz[0], r*u._xyz[1]) ;
	}


	// -----------------------------------------------------------------------------
	//! Norm
	template<class T> T T_Vector2<T>::norm () const
	{
		return sqrt( _xyz[0]*_xyz[0] + _xyz[1]*_xyz[1] );
	}

	// -----------------------------------------------------------------------------
	//! Square norm (self dot product)
	template<class T> T T_Vector2<T>::squareNorm () const
	{
		return _xyz[0]*_xyz[0] + _xyz[1]*_xyz[1] ;
	}

	// -----------------------------------------------------------------------------
	//! Infinite norm 
	template<class T> T T_Vector2<T>::infNorm() const
	{
		return std::max(fabs(_xyz[0]),fabs(_xyz[1])) ;
	}


	// -----------------------------------------------------------------------------
	//! Out stream override: prints the 3 vector components
	template<class T> std::ostream& operator<< (std::ostream& out,const T_Vector2<T>& u)
	{
		out << u[0] << " " << u[1] ;
		return ( out );
	}

	template<class T> T_Vector2<T> min(const T_Vector2<T>& v1,const T_Vector2<T>& v2) 
	{
		return T_Vector2<T>(std::min(v1[0],v2[0]),std::min(v1[1],v2[1])) ;
	}

	template<class T> T_Vector2<T> max(const T_Vector2<T>& v1,const T_Vector2<T>& v2) 
	{
		return T_Vector2<T>(std::max(v1[0],v2[0]),std::max(v1[1],v2[1])) ;
	}
}

#endif // _VECTOR3_H
